#include <Arduino.h>
#include <pico.h>
#include <hardware/pwm.h>
#include <Keyboard.h>
#include <FastLED.h>

#define PORT 2
#define PORT2 3
#define THRESHOLD 10
#define LEDPORT 5

CRGB leds[2];

class HallButton
{
public:
  uint64_t firstPulseStart;
  uint8_t counter = 0;
  uint64_t firstPulseLength[3]; // value of button
  int64_t value;
  uint64_t _median;
  int64_t maxValue;
  bool isFirstClick = true;
  bool isDataReadyToRead = false;
  bool isButtonPressed = false;
  uint64_t nextTogglePoint = 0; // value when to toggle state of button
  size_t ledId;

  HallButton(uint8_t gpioNum, char buttonOnKeyboard, size_t ledId)
  {
    this->gpioNum = gpioNum;
    this->buttonOnKeyboard = buttonOnKeyboard;
    this->ledId = ledId;
  }

  void HandleISR(bool gpioState, uint64_t currentTime)
  {
    if (gpioState)
    {
      firstPulseLength[counter] = currentTime - firstPulseStart;
      isDataReadyToRead = true;
      if (++counter >= 3)
        counter = 0;
    }
    else
    {
      firstPulseStart = currentTime;
    }
  }
  void HandleButton()
  {
    if (!isDataReadyToRead)
      return;
    _median = median(firstPulseLength[0], firstPulseLength[1], firstPulseLength[2]);
    value = filtered((float)_median);
    if (value > maxValue)
      if (isFirstClick)
        maxValue = value;
      else
        value = maxValue;

    if (isButtonPressed)
    {
      if (value < nextTogglePoint)
      {
        Release();
        if (isFirstClick)
          isFirstClick = false;
      }
      if ((int64_t)value - (int64_t)nextTogglePoint > THRESHOLD)
        nextTogglePoint = value - THRESHOLD;
    }
    else
    {
      if (value >= nextTogglePoint)
      {
        Press();
      }
      if ((int64_t)nextTogglePoint - (int64_t)value > THRESHOLD)
        nextTogglePoint = value + THRESHOLD;
    }
    isDataReadyToRead = false;
  }
  uint64_t median(uint64_t a, uint64_t b, uint64_t c)
  {
    return (a < b) ? ((b < c) ? b : ((c < a) ? a : c)) : ((a < c) ? a : ((c < b) ? b : c));
  }
  float filtered(float value)
  {
    float _kalman_gain, _current_estimate;
    _kalman_gain = _err_estimate / (_err_estimate + _err_measure);
    _current_estimate = _last_estimate + _kalman_gain * (value - _last_estimate);
    _err_estimate = (1.0 - _kalman_gain) * _err_estimate + fabs(_last_estimate - _current_estimate) * _q;
    _last_estimate = _current_estimate;
    return _current_estimate;
  }

private:
  float _err_measure = 15;
  float _err_estimate = 15;
  float _q = 0.1;
  float _last_estimate = 0.0;
  uint8_t gpioNum;
  char buttonOnKeyboard;
  void Press()
  {
    if (isButtonPressed)
      return;
    Keyboard.press(buttonOnKeyboard);
    isButtonPressed = true;
    LEDTick(isButtonPressed, ledId);
  }
  void Release()
  {
    if (!isButtonPressed)
      return;
    Keyboard.release(buttonOnKeyboard);
    isButtonPressed = false;
    LEDTick(isButtonPressed, ledId);
  }
  static void LEDTick(bool status, size_t ledId)
  {
    leds[ledId] = status ? CRGB::Red : CRGB::Black;
    FastLED.show();
  }
};

HallButton btn[2] = {HallButton(PORT, 't', 1), HallButton(PORT2, 'y', 0)};

void isr(uint gpio, uint32_t events)
{
  uint64_t currentTime = time_us_64();
  bool state = events == 8;
  switch (gpio)
  {
  case PORT:
    btn[0].HandleISR(state, currentTime);
    break;
  case PORT2:
    btn[1].HandleISR(state, currentTime);
    break;
  }
}

void setup()
{
  gpio_init(PORT);
  gpio_set_dir(PORT, false);
  gpio_pull_up(PORT);
  gpio_set_irq_enabled_with_callback(PORT, GPIO_IRQ_EDGE_RISE | GPIO_IRQ_EDGE_FALL, true, isr);
  gpio_init(PORT2);
  gpio_set_dir(PORT2, false);
  gpio_pull_up(PORT2);
  gpio_set_irq_enabled_with_callback(PORT2, GPIO_IRQ_EDGE_RISE | GPIO_IRQ_EDGE_FALL, true, isr);
  // Serial.begin(9600);
  Keyboard.begin();
  FastLED.addLeds<NEOPIXEL, LEDPORT>(leds, 2);
}

void printDebugInfo()
{
  static uint32_t lastTime = 0;
  uint32_t time = millis();
  if (time - lastTime >= 1)
  {
    Serial.printf("%d,%d,%llu,%llu,%i,", btn[0].isFirstClick, btn[0].maxValue, btn[0].value, btn[0].nextTogglePoint, btn[0].isButtonPressed);
    Serial.printf("%d,%d,%llu,%llu,%i\n", btn[1].isFirstClick, btn[1].maxValue, btn[1].value, btn[1].nextTogglePoint, btn[1].isButtonPressed);
    lastTime = time;
  }
}
void loop()
{
  btn[0].HandleButton();
  btn[1].HandleButton();
}